@@ -19,6 +19,7 @@ import json
 import apt_pkg
 
 from UbuntuDrivers import kerneldetection
+from functools import cmp_to_key
 
 system_architecture = ''
 lookup_cache = {}
@@ -48,7 +49,7 @@ class NvidiaPkgNameInfo(object):
         # Obsolete naming such as nvidia-340
         if match:
             self._obsolete_name_scheme = True
-            self._major_ver = match.group(1)
+            self._major_ver = int(match.group(1))
             self._flavour = self._major_ver
             self.is_valid = True
             return
@@ -63,6 +64,20 @@ class NvidiaPkgNameInfo(object):
             self._flavour = '%s%s%s' % (match.group(1),
                                         '-server' if self._server else '',
                                         '-open' if self._open else '')
+            self._major_ver = int(match.group(1))
+            self.is_valid = True
+
+        # Recent naming such as nvidia-headless[-no-dkms]-525 for gpgpu
+        pattern = re.compile('nvidia-headless(.*)-([0-9]+)(.*)')
+        match = pattern.match(name)
+
+        if match:
+            self._server = match.group(0).find('-server') != -1
+            self._open = match.group(0).find('-open') != -1
+            self._flavour = '%s%s%s' % (match.group(2),
+                                        '-server' if self._server else '',
+                                        '-open' if self._open else '')
+            self._major_ver = match.group(2)
             self.is_valid = True
 
     def has_obsolete_name_scheme(self):
@@ -357,13 +372,33 @@ def _is_package_from_distro(apt_cache, pkg):
     candidate = depcache.get_candidate_ver(pkg)
     if candidate is None:
         return False
-
     try:
         return (candidate.file_list[0][0].origin == 'Ubuntu')
     except KeyError:
         return False
 
 
+def _pkg_get_open_preference(apt_cache, pkg):
+    '''Determine if -open package is prefered from apt Package object'''
+    depcache = apt_pkg.DepCache(apt_cache)
+    candidate = depcache.get_candidate_ver(pkg)
+    records = apt_pkg.PackageRecords(apt_cache)
+    records.lookup(candidate.file_list[0])
+
+    try:
+        preference = records['Prefer-Variant']
+    except (KeyError, AttributeError):
+        logging.debug('_pkg_get_open_preference %s: package has Prefer-Open header, cannot determine', pkg.name)
+        return False
+
+    if preference not in ('Open', 'Closed'):
+        logging.debug('_pkg_get_open_preference %s: package has invalid Prefer-Open %s'
+                      'header, cannot determine preference', pkg.name, preference)
+        return None
+
+    return preference
+
+
 def _pkg_get_module(apt_cache, pkg):
     '''Determine module name from apt Package object'''
     depcache = apt_pkg.DepCache(apt_cache)
@@ -507,10 +542,10 @@ def _get_db_name(syspath, alias):
     Values are None if unknown.
     '''
     try:
-        out = subprocess.check_output(['udevadm', 'hwdb', '--test=' + alias],
+        out = subprocess.check_output(['systemd-hwdb', 'query', alias],
                                       universal_newlines=True)
     except (OSError, subprocess.CalledProcessError) as e:
-        logging.debug('_get_db_name(%s, %s): udevadm hwdb failed: %s', syspath, alias, str(e))
+        logging.debug('_get_db_name(%s, %s): systemd-hwdb failed: %s', syspath, alias, str(e))
         return (None, None)
 
     logging.debug('_get_db_name: output\n%s\n', out)
@@ -529,6 +564,24 @@ def _get_db_name(syspath, alias):
     return (vendor, model)
 
 
+def _is_open_prefered(apt_cache, pkg):
+    if not pkg.name.startswith('nvidia-'):
+        return False
+
+    preference = _pkg_get_open_preference(apt_cache, pkg)
+    if preference is None:
+        nvidia_info = NvidiaPkgNameInfo(pkg.name)
+        if nvidia_info.get_major_version() >= 560:
+            logging.debug('_is_open_prefered(%s): True', pkg.name)
+            return True
+    elif preference == 'Open':
+        logging.debug('_is_open_prefered(%s): True', pkg.name)
+        return True
+
+    logging.debug('_is_open_prefered(%s): False', pkg.name)
+    return False
+
+
 def set_nvidia_kms(value):
     '''Set KMS on or off for NVIDIA'''
     nvidia_kms_file = '/lib/modprobe.d/nvidia-kms.conf'
@@ -599,7 +652,8 @@ def system_driver_packages(apt_cache=None, sys_path=None, freeonly=False, includ
                     'free': _is_package_free(apt_cache, p),
                     'from_distro': _is_package_from_distro(apt_cache, p),
                     'support': _pkg_get_support(apt_cache, p),
-                    'runtimepm': _is_runtimepm_supported(apt_cache, p, alias)
+                    'runtimepm': _is_runtimepm_supported(apt_cache, p, alias),
+                    'open_preferred': _is_open_prefered(apt_cache, p)
                 }
             (vendor, model) = _get_db_name(syspath, alias)
             if vendor is not None:
@@ -775,6 +829,7 @@ def system_device_specific_metapackages(apt_cache=None, sys_path=None, include_o
                     'from_distro': _is_package_from_distro(apt_cache, p),
                     'recommended': True,
                     'support': _pkg_get_support(apt_cache, p),
+                    'open_preferred': _is_open_prefered(apt_cache, p),
                 }
 
     return packages
@@ -836,6 +891,7 @@ def system_gpgpu_driver_packages(apt_cache=None, sys_path=None):
                         'free': _is_package_free(apt_cache, p),
                         'from_distro': _is_package_from_distro(apt_cache, p),
                         'support': _pkg_get_support(apt_cache, p),
+                        'open_preferred': _is_open_prefered(apt_cache, p),
                     }
                 if vendor is not None:
                     packages[p.name]['vendor'] = vendor
@@ -954,12 +1010,14 @@ def system_device_drivers(apt_cache=None, sys_path=None, freeonly=False):
     return result
 
 
-def get_desktop_package_list(apt_cache, sys_path=None, free_only=False, include_oem=True, driver_string=''):
+def get_desktop_package_list(
+        apt_cache, sys_path=None, free_only=False, include_oem=True,
+        driver_string='', include_dkms=False):
     '''Return the list of packages that should be installed'''
     packages = system_driver_packages(
         apt_cache, sys_path, freeonly=free_only,
         include_oem=include_oem)
-    packages = auto_install_filter(packages, driver_string)
+    packages = auto_install_filter(packages, driver_string, get_recommended=False)
     if not packages:
         logging.debug('No drivers found for installation.')
         return packages
@@ -969,13 +1027,15 @@ def get_desktop_package_list(apt_cache, sys_path=None, free_only=False, include_
 
     # ignore packages which are already installed
     to_install = []
-    for p in packages:
+    for p, _ in sorted(packages.items(),
+                       key=cmp_to_key(lambda left, right: _cmp_gfx_alternatives(left[0], right[0])),
+                       reverse=True):
         package_obj = apt_cache[p]
         if not package_obj.current_ver:
-            to_install.append(p)
 
             candidate = depcache.get_candidate_ver(package_obj)
             records.lookup(candidate.file_list[0])
+            to_install.append(p)
 
             # See if runtimepm is supported
             if records['runtimepm']:
@@ -992,6 +1052,10 @@ def get_desktop_package_list(apt_cache, sys_path=None, free_only=False, include_
             try:
                 modules_package = get_linux_modules_metapackage(apt_cache, p)
                 if modules_package and not apt_cache[modules_package].current_ver:
+                    if not include_dkms and "dkms" in modules_package:
+                        to_install.remove(p)
+                        continue
+                    to_install.remove(p)
                     to_install.append(modules_package)
 
                     lrm_meta = get_userspace_lrm_meta(apt_cache, p)
@@ -999,6 +1063,7 @@ def get_desktop_package_list(apt_cache, sys_path=None, free_only=False, include_
                         # Add the lrm meta and drop the non lrm one
                         to_install.append(lrm_meta)
                         to_install.remove(p)
+                    break
             except KeyError:
                 pass
 
@@ -1069,7 +1134,7 @@ def _process_driver_string(string):
     return driver
 
 
-def gpgpu_install_filter(packages, drivers_str):
+def gpgpu_install_filter(packages, drivers_str, get_recommended=True):
     drivers = []
     allow = []
     result = {}
@@ -1162,14 +1227,17 @@ def gpgpu_install_filter(packages, drivers_str):
                     result[p] = packages[p]
                 else:
                     # print('before recommended: %s' % packages[p])
-                    if packages[p].get('recommended'):
+                    if get_recommended:
+                        if packages[p].get('recommended'):
+                            result[p] = packages[p]
+                    else:
                         result[p] = packages[p]
                         # print('Found "recommended" flavour in %s' % (packages[p]))
                 break
     return result
 
 
-def auto_install_filter(packages, drivers_str=''):
+def auto_install_filter(packages, drivers_str='', get_recommended=True):
     '''Get packages which are appropriate for automatic installation.
 
     Return the subset of the given list of packages which are appropriate for
@@ -1180,7 +1248,7 @@ def auto_install_filter(packages, drivers_str=''):
     '''
     # any package which matches any of those globs will be accepted
     whitelist = ['bcmwl*', 'pvr-omap*', 'virtualbox-guest*', 'nvidia-*',
-                 'open-vm-tools*', 'oem-*-meta']
+                 'open-vm-tools*', 'hwe-*-meta', 'oem-*-meta', 'broadcom-sta-dkms']
 
     # If users specify a driver, use gpgpu_install_filter()
     if drivers_str:
@@ -1193,7 +1261,10 @@ def auto_install_filter(packages, drivers_str=''):
 
     result = {}
     for p in allow:
-        if 'recommended' not in packages[p] or packages[p]['recommended']:
+        if get_recommended:
+            if 'recommended' not in packages[p] or packages[p]['recommended']:
+                result[p] = packages[p]
+        else:
             result[p] = packages[p]
     return result
 
@@ -1267,66 +1338,140 @@ def _pkg_support_from_cache(x):
     return None
 
 
+def _pkg_open_preferred_from_cache(x):
+    '''Look up driver and return if open package is preferred'''
+    if lookup_cache.get(x):
+        return lookup_cache.get(x).get('open_preferred')
+    return False
+
+
+def _get_fit_level(x):
+    '''decide how well the package fits desktop environment'''
+
+    # for desktop non-server packages are prefered to -server ones
+    # if a package preferes open variants, these come on top
+
+    if _pkg_open_preferred_from_cache(x):
+        if re.search(r'(\d{3})-open$', x):
+            return 3
+
+        if re.search(r'(\d{3})$', x):
+            return 2
+
+        if re.search(r'(\d{3})-server-open$', x):
+            return 1
+
+        if re.search(r'(\d{3})-server$', x):
+            return 0
+    else:
+        if re.search(r'(\d{3})$', x):
+            return 3
+
+        if re.search(r'(\d{3})-server$', x):
+            return 2
+
+        if re.search(r'(\d{3})-open$', x):
+            return 1
+
+        if re.search(r'(\d{3})-server-open$', x):
+            return 0
+    return 0
+
+
+def _get_fit_level_gpgpu(x):
+    '''decide how well the package fits server environment'''
+
+    # for server environment -server packages are prefered to non-server ones
+    # if a package preferes open variants, these come on top
+
+    if _pkg_open_preferred_from_cache(x):
+        if re.search(r'(\d{3})-server-open$', x):
+            return 3
+
+        if re.search(r'(\d{3})-server$', x):
+            return 2
+
+        if re.search(r'(\d{3})-open$', x):
+            return 1
+
+        if re.search(r'(\d{3})$', x):
+            return 0
+    else:
+        if re.search(r'(\d{3})-server$', x):
+            return 3
+
+        if re.search(r'(\d{3})$', x):
+            return 2
+
+        if re.search(r'(\d{3})-server-open$', x):
+            return 1
+
+        if re.search(r'(\d{3})-open$', x):
+            return 0
+    return 0
+
+
 def _cmp_gfx_alternatives(x, y):
     '''Compare two graphics driver names in terms of preference. (desktop)
 
-    -open always sorts after non-open.
+    -open by default sorts after non-open, unless a driver preferes open variant
     -server always sorts after non-server.
     LTSB (Long Term Support Branch) always sorts before NFB (New Feature Branch).
     Legacy always sorts before Beta.
     '''
 
-    if x.endswith('-open') and not y.endswith('-open'):
-        return -1
-    if not x.endswith('-open') and y.endswith('-open'):
-        return 1
-
-    if x.endswith('-server') and not y.endswith('-server'):
-        return -1
-    if not x.endswith('-server') and y.endswith('-server'):
-        return 1
-
-    preferred_support = ['PB', 'LTSB']
-
-    x_score = 0
-    y_score = 0
-
-    x_support = _pkg_support_from_cache(x)
-    y_support = _pkg_support_from_cache(y)
-
-    if x_support in preferred_support:
-        x_score += 100
-
-    if y_support in preferred_support:
-        y_score += 100
-
-    if x > y:
-        x_score += 1
-    elif x < y:
-        y_score += 1
-
-    if ((x_score >= 100) or (y_score >= 100)):
-        if x_score > y_score:
+    if x.startswith('nvidia-') and y.startswith('nvidia-'):
+        if _get_fit_level(x) > _get_fit_level(y):
             return 1
-        elif x_score < y_score:
+        if _get_fit_level(x) < _get_fit_level(y):
             return -1
 
-    if _pkg_support_from_cache(x) == 'PB' and _pkg_support_from_cache(y) != 'PB':
-        return 1
-    if _pkg_support_from_cache(x) != 'PB' and _pkg_support_from_cache(y) == 'PB':
-        return -1
-    if _pkg_support_from_cache(x) == 'LTSB' and _pkg_support_from_cache(y) != 'LTSB':
-        return 1
-    if _pkg_support_from_cache(x) != 'LTSB' and _pkg_support_from_cache(y) == 'LTSB':
-        return -1
-    if _pkg_support_from_cache(x) == 'Legacy' and _pkg_support_from_cache(y) != 'Legacy':
-        return -1
-    if _pkg_support_from_cache(x) != 'Legacy' and _pkg_support_from_cache(y) == 'Legacy':
-        return 1
-    if _pkg_support_from_cache(x) == 'Beta' and _pkg_support_from_cache(y) != 'Beta':
-        return -1
-    if _pkg_support_from_cache(x) != 'Beta' and _pkg_support_from_cache(y) == 'Beta':
-        return 1
+        preferred_support = ['PB', 'LTSB']
+
+        x_score = 0
+        y_score = 0
+
+        x_support = _pkg_support_from_cache(x)
+        y_support = _pkg_support_from_cache(y)
+
+        if x_support in preferred_support:
+            x_score += 100
+
+        if y_support in preferred_support:
+            y_score += 100
+
+        if x > y:
+            x_score += 1
+        elif x < y:
+            y_score += 1
+
+        if ((x_score >= 100) or (y_score >= 100)):
+            if x_score > y_score:
+                return 1
+            elif x_score < y_score:
+                return -1
+
+        if _pkg_support_from_cache(x) == 'PB' and _pkg_support_from_cache(y) != 'PB':
+            return 1
+        if _pkg_support_from_cache(x) != 'PB' and _pkg_support_from_cache(y) == 'PB':
+            return -1
+        if _pkg_support_from_cache(x) == 'LTSB' and _pkg_support_from_cache(y) != 'LTSB':
+            return 1
+        if _pkg_support_from_cache(x) != 'LTSB' and _pkg_support_from_cache(y) == 'LTSB':
+            return -1
+        if _pkg_support_from_cache(x) == 'NFB' and _pkg_support_from_cache(y) != 'NFB':
+            return -1
+        if _pkg_support_from_cache(x) != 'NFB' and _pkg_support_from_cache(y) == 'NFB':
+            return 1
+        if _pkg_support_from_cache(x) == 'Legacy' and _pkg_support_from_cache(y) != 'Legacy':
+            return -1
+        if _pkg_support_from_cache(x) != 'Legacy' and _pkg_support_from_cache(y) == 'Legacy':
+            return 1
+        if _pkg_support_from_cache(x) == 'Beta' and _pkg_support_from_cache(y) != 'Beta':
+            return -1
+        if _pkg_support_from_cache(x) != 'Beta' and _pkg_support_from_cache(y) == 'Beta':
+            return 1
+
     if x < y:
         return -1
     if x > y:
@@ -1338,37 +1483,40 @@ def _cmp_gfx_alternatives(x, y):
 def _cmp_gfx_alternatives_gpgpu(x, y):
     '''Compare two graphics driver names in terms of preference. (server)
 
-    -open always sorts after non-open.
+    -open by default sorts after non-open, unless a driver preferes open variant
     -server always sorts before non-server.
+    non server package always sorts after server package
     LTSB (Long Term Support Branch) always sorts before NFB (New Feature Branch).
     Legacy always sorts before Beta.
     '''
-    if x.endswith('-open') and not y.endswith('-open'):
-        return -1
-    if not x.endswith('-open') and y.endswith('-open'):
-        return 1
 
-    if x.endswith('-server') and not y.endswith('-server'):
-        return 1
-    if not x.endswith('-server') and y.endswith('-server'):
-        return -1
+    if x.startswith('nvidia-') and y.startswith('nvidia-'):
+        if _get_fit_level_gpgpu(x) > _get_fit_level_gpgpu(y):
+            return 1
+        if _get_fit_level_gpgpu(x) < _get_fit_level_gpgpu(y):
+            return -1
+
+        if _pkg_support_from_cache(x) == 'PB' and _pkg_support_from_cache(y) != 'PB':
+            return 1
+        if _pkg_support_from_cache(x) != 'PB' and _pkg_support_from_cache(y) == 'PB':
+            return -1
+        if _pkg_support_from_cache(x) == 'LTSB' and _pkg_support_from_cache(y) != 'LTSB':
+            return 1
+        if _pkg_support_from_cache(x) != 'LTSB' and _pkg_support_from_cache(y) == 'LTSB':
+            return -1
+        if _pkg_support_from_cache(x) == 'Legacy' and _pkg_support_from_cache(y) != 'Legacy':
+            return -1
+        if _pkg_support_from_cache(x) != 'Legacy' and _pkg_support_from_cache(y) == 'Legacy':
+            return 1
+        if _pkg_support_from_cache(x) == 'NFB' and _pkg_support_from_cache(y) != 'NFB':
+            return -1
+        if _pkg_support_from_cache(x) != 'NFB' and _pkg_support_from_cache(y) == 'NFB':
+            return 1
+        if _pkg_support_from_cache(x) == 'Beta' and _pkg_support_from_cache(y) != 'Beta':
+            return -1
+        if _pkg_support_from_cache(x) != 'Beta' and _pkg_support_from_cache(y) == 'Beta':
+            return 1
 
-    if _pkg_support_from_cache(x) == 'PB' and _pkg_support_from_cache(y) != 'PB':
-        return 1
-    if _pkg_support_from_cache(x) != 'PB' and _pkg_support_from_cache(y) == 'PB':
-        return -1
-    if _pkg_support_from_cache(x) == 'LTSB' and _pkg_support_from_cache(y) != 'LTSB':
-        return 1
-    if _pkg_support_from_cache(x) != 'LTSB' and _pkg_support_from_cache(y) == 'LTSB':
-        return -1
-    if _pkg_support_from_cache(x) == 'Legacy' and _pkg_support_from_cache(y) != 'Legacy':
-        return -1
-    if _pkg_support_from_cache(x) != 'Legacy' and _pkg_support_from_cache(y) == 'Legacy':
-        return 1
-    if _pkg_support_from_cache(x) == 'Beta' and _pkg_support_from_cache(y) != 'Beta':
-        return -1
-    if _pkg_support_from_cache(x) != 'Beta' and _pkg_support_from_cache(y) == 'Beta':
-        return 1
     if x < y:
         return -1
     if x > y:
@@ -1518,8 +1666,11 @@ def get_linux_modules_metapackage(apt_cache, candidate):
                         if dep.parent_pkg.name.startswith('linux-modules-nvidia-')]
 
         pick = ''
-        modules_candidate = 'linux-modules-nvidia-%s-%s' % (candidate_flavour,
-                                                            get_linux_image(apt_cache).replace('linux-image-', ''))
+        candidate_suffix = get_linux_image(apt_cache).replace('linux-image-', '')
+        if candidate_suffix == "virtual":
+            candidate_suffix = "generic"
+        modules_candidate = 'linux-modules-nvidia-%s-%s' % (candidate_flavour, candidate_suffix)
+
         for dep in reverse_deps:
             if dep == modules_candidate:
                 pick = dep
